@@ -16,89 +16,28 @@
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 import argparse
+import asyncio
 import json
 import logging
 import pathlib
 import random
-import socket
+import ssl
 import time
 import typing
 
-import paho.mqtt.client
+import aiomqtt
 import wireless_sensor
 
 import wireless_sensor_mqtt._homeassistant
 
 _MQTT_DEFAULT_PORT = 1883
 _MQTT_DEFAULT_TLS_PORT = 8883
-_MQTT_PUBLISH_TIMEOUT_SECONDS = 16
-_MQTT_PUBLISH_STATUS_POLL_INTERVAL_SECONDS = 1
 _MEASUREMENT_MOCKS_COUNT = 3
 _MEASUREMENT_MOCKS_INTERVAL_SECONDS = 8
 
 _LOGGER = logging.getLogger(__name__)
 
 
-def _mqtt_on_connect(
-    mqtt_client: paho.mqtt.client.Client,
-    userdata: None,
-    flags: typing.Dict,
-    return_code: int,
-) -> None:
-    # pylint: disable=unused-argument; callback
-    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.0/src/paho/mqtt/client.py#L441
-    assert return_code == 0, return_code  # connection accepted
-    mqtt_broker_host, mqtt_broker_port = mqtt_client.socket().getpeername()
-    _LOGGER.debug("connected to MQTT broker %s:%d", mqtt_broker_host, mqtt_broker_port)
-
-
-def _init_mqtt_client(
-    # *, SyntaxError on python3.5
-    host: str,
-    port: int,
-    username: typing.Optional[str],
-    password: typing.Optional[str],
-    disable_tls: bool,
-) -> paho.mqtt.client.Client:
-    # https://pypi.org/project/paho-mqtt/
-    client = paho.mqtt.client.Client()
-    client.on_connect = _mqtt_on_connect
-    if not disable_tls:
-        client.tls_set(ca_certs=None)  # enable tls trusting default system certs
-    _LOGGER.info(
-        "connecting to MQTT broker %s:%d (TLS %s)",
-        host,
-        port,
-        "disabled" if disable_tls else "enabled",
-    )
-    if username:
-        client.username_pw_set(username=username, password=password)
-    elif password:
-        raise ValueError("Missing MQTT username")
-    client.connect(host=host, port=port)
-    return client
-
-
-def _mqtt_attempt_reconnect(client: paho.mqtt.client.Client) -> None:
-    """
-    > Calling connect() or reconnect() will cause the messages to be resent.
-    > Use reinitialise() to reset a client to its original state.
-    https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L530
-    """
-    _LOGGER.debug("reconnecting to MQTT broker")
-    try:
-        return_code = client.reconnect()
-        if return_code != paho.mqtt.client.MQTT_ERR_SUCCESS:
-            raise RuntimeError(
-                "Client.reconnect() returned expected return code {}".format(
-                    return_code
-                )
-            )
-    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L1805
-    except (socket.error, OSError, RuntimeError):
-        _LOGGER.error("failed to reconnect to MQTT broker", exc_info=True)
-
-
 def _mock_measurement() -> wireless_sensor.Measurement:
     time.sleep(_MEASUREMENT_MOCKS_INTERVAL_SECONDS)
     return wireless_sensor.Measurement(
@@ -108,40 +47,9 @@ def _mock_measurement() -> wireless_sensor.Measurement:
     )
 
 
-def _mqtt_publish(
-    *, client: paho.mqtt.client.Client, topic: str, payload: str, **kwargs
-) -> None:
-    _LOGGER.debug("publishing mqtt msg: topic=%s payload=%s", topic, payload)
-    msg_info = client.publish(
-        topic=topic, payload=payload, **kwargs
-    )  # type: paho.mqtt.client.MQTTMessageInfo
-    # MQTTMessageInfo.wait_for_publish() calls threading.Condition.wait() without timeout
-    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L338
-    poll_start_time = time.time()
-    while (
-        not msg_info.is_published()
-        and (time.time() - poll_start_time) < _MQTT_PUBLISH_TIMEOUT_SECONDS
-    ):
-        if msg_info.rc == paho.mqtt.client.MQTT_ERR_NO_CONN:
-            _mqtt_attempt_reconnect(client)
-        time.sleep(_MQTT_PUBLISH_STATUS_POLL_INTERVAL_SECONDS)
-    # https://github.com/eclipse/paho.mqtt.python/blob/v1.5.1/src/paho/mqtt/client.py#L147
-    if msg_info.rc != paho.mqtt.client.MQTT_ERR_SUCCESS:
-        _LOGGER.error(
-            "failed to publish on topic %s (return code %d)", topic, msg_info.rc
-        )
-    elif not msg_info.is_published():
-        _LOGGER.warning(
-            "reached timeout of %d seconds"
-            " while waiting for MQTT message on topic %s to get published",
-            _MQTT_PUBLISH_TIMEOUT_SECONDS,
-            topic,
-        )
-
-
-def _publish_homeassistant_discovery_config(
-    # *, SyntaxError on python3.5
-    mqtt_client: paho.mqtt.client.Client,
+async def _publish_homeassistant_discovery_config(
+    *,
+    mqtt_client: aiomqtt.Client,
     homeassistant_discovery_prefix: str,
     homeassistant_node_id: str,
     temperature_topic: str,
@@ -157,7 +65,7 @@ def _publish_homeassistant_discovery_config(
         # > voluptuous.error.MultipleInvalid: Device must have at least one id
         # > entifying value in 'identifiers' and/or 'connections'
         # > for dictionary value @ data['device']
-        "identifiers": ["FT017TH/{}".format(homeassistant_node_id)],
+        "identifiers": [f"FT017TH/{homeassistant_node_id}"],
         "model": "FT017TH",
     }
     for object_id, device_class, state_topic, unit, name_suffix in zip(
@@ -192,31 +100,33 @@ def _publish_homeassistant_discovery_config(
             "unique_id": unique_id,
             "device_class": device_class,
             # friendly_name & template for default entity_id
-            "name": "{} {}".format(homeassistant_node_id, name_suffix),
+            "name": f"{homeassistant_node_id} {name_suffix}",
             "state_topic": state_topic,
             "unit_of_measurement": unit,
             "expire_after": 60 * 10,  # seconds
             "device": device_attrs,
         }
-        _mqtt_publish(
-            client=mqtt_client,
-            topic=discovery_topic,
-            payload=json.dumps(config),
-            retain=True,
+        await mqtt_client.publish(
+            topic=discovery_topic, payload=json.dumps(config), retain=True
         )
 
 
-def _measurement_iter(
-    mock_measurements: bool, unlock_spi_device: bool
-) -> typing.Iterator[wireless_sensor.Measurement]:
+async def _measurement_iter(
+    mock_measurements: bool, gdo0_gpio_line_name: bytes, unlock_spi_device: bool
+) -> typing.AsyncIterator[wireless_sensor.Measurement]:
     if mock_measurements:
         logging.warning("publishing %d mocked measurements", _MEASUREMENT_MOCKS_COUNT)
-        return map(lambda _: _mock_measurement(), range(_MEASUREMENT_MOCKS_COUNT))
-    return wireless_sensor.FT017TH(unlock_spi_device=unlock_spi_device).receive()
+        for _ in range(_MEASUREMENT_MOCKS_COUNT):
+            yield _mock_measurement()
+    else:
+        async for measurement in wireless_sensor.FT017TH(
+            gdo0_gpio_line_name=gdo0_gpio_line_name, unlock_spi_device=unlock_spi_device
+        ).receive(timeout_seconds=3600):
+            yield measurement
 
 
-def _run(
-    # *, SyntaxError on python3.5
+async def _run(  # pylint: disable=too-many-locals
+    *,
     mqtt_host: str,
     mqtt_port: int,
     mqtt_disable_tls: bool,
@@ -226,50 +136,62 @@ def _run(
     homeassistant_discovery_prefix: str,
     homeassistant_node_id: str,
     mock_measurements: bool,
+    gdo0_gpio_line_name: bytes,
     unlock_spi_device: bool,
 ) -> None:
     # pylint: disable=too-many-arguments
-    # https://pypi.org/project/paho-mqtt/
-    mqtt_client = _init_mqtt_client(
-        host=mqtt_host,
+    _LOGGER.info(
+        "connecting to MQTT broker %s:%d (TLS %s)",
+        mqtt_host,
+        mqtt_port,
+        "disabled" if mqtt_disable_tls else "enabled",
+    )
+    if mqtt_password and not mqtt_username:
+        raise ValueError("Missing MQTT username")
+    async with aiomqtt.Client(
+        tls_context=None if mqtt_disable_tls else ssl.create_default_context(),
+        hostname=mqtt_host,
         port=mqtt_port,
-        disable_tls=mqtt_disable_tls,
         username=mqtt_username,
         password=mqtt_password,
-    )
-    temperature_topic = mqtt_topic_prefix + "/temperature-degrees-celsius"
-    humidity_topic = mqtt_topic_prefix + "/relative-humidity-percent"
-    _LOGGER.debug(
-        "publishing measurements on topics %r and %r", temperature_topic, humidity_topic
-    )
-    homeassistant_discover_config_published = False
-    for measurement in _measurement_iter(
-        mock_measurements=mock_measurements, unlock_spi_device=unlock_spi_device
-    ):
-        _LOGGER.debug("received %s", measurement)
-        if not homeassistant_discover_config_published:
-            _publish_homeassistant_discovery_config(
-                mqtt_client=mqtt_client,
-                homeassistant_discovery_prefix=homeassistant_discovery_prefix,
-                homeassistant_node_id=homeassistant_node_id,
-                temperature_topic=temperature_topic,
-                humidity_topic=humidity_topic,
+    ) as mqtt_client:
+        _LOGGER.debug("connected to MQTT broker %s:%d", mqtt_host, mqtt_port)
+        temperature_topic = mqtt_topic_prefix + "/temperature-degrees-celsius"
+        humidity_topic = mqtt_topic_prefix + "/relative-humidity-percent"
+        _LOGGER.debug(
+            "publishing measurements on topics %r and %r",
+            temperature_topic,
+            humidity_topic,
+        )
+        homeassistant_discover_config_published = False
+        async for measurement in _measurement_iter(
+            mock_measurements=mock_measurements,
+            gdo0_gpio_line_name=gdo0_gpio_line_name,
+            unlock_spi_device=unlock_spi_device,
+        ):
+            _LOGGER.debug("received %s", measurement)
+            if not homeassistant_discover_config_published:
+                await _publish_homeassistant_discovery_config(
+                    mqtt_client=mqtt_client,
+                    homeassistant_discovery_prefix=homeassistant_discovery_prefix,
+                    homeassistant_node_id=homeassistant_node_id,
+                    temperature_topic=temperature_topic,
+                    humidity_topic=humidity_topic,
+                )
+                homeassistant_discover_config_published = True
+            await mqtt_client.publish(
+                topic=temperature_topic,
+                payload=f"{measurement.temperature_degrees_celsius:.02f}",
+                retain=False,
             )
-            homeassistant_discover_config_published = True
-        _mqtt_publish(
-            client=mqtt_client,
-            topic=temperature_topic,
-            payload="{:.02f}".format(measurement.temperature_degrees_celsius),
-            retain=False,
-        )
-        _mqtt_publish(
-            client=mqtt_client,
-            topic=humidity_topic,
-            # > unit_of_measurement: '%'
-            # https://www.home-assistant.io/integrations/sensor.mqtt/#temperature-and-humidity-sensors
-            payload="{:.02f}".format(measurement.relative_humidity * 100),
-            retain=False,
-        )
+            await mqtt_client.publish(
+                topic=humidity_topic,
+                # > unit_of_measurement: '%'
+                # https://www.home-assistant.io/integrations/sensor.mqtt/#temperature-and-humidity-sensors
+                payload=f"{(measurement.relative_humidity * 100):.02f}",
+                retain=False,
+            )
+        raise RuntimeError("timeout waiting for packet")
 
 
 def _main() -> None:
@@ -281,9 +203,7 @@ def _main() -> None:
     argparser.add_argument(
         "--mqtt-port",
         type=int,
-        help="default {} ({} with --mqtt-disable-tls)".format(
-            _MQTT_DEFAULT_TLS_PORT, _MQTT_DEFAULT_PORT
-        ),
+        help=f"default {_MQTT_DEFAULT_TLS_PORT} ({_MQTT_DEFAULT_PORT} with --mqtt-disable-tls)",
     )
     argparser.add_argument("--mqtt-username", type=str)
     argparser.add_argument("--mqtt-disable-tls", action="store_true")
@@ -318,6 +238,16 @@ def _main() -> None:
         action="store_true",
         help="publish random values to test MQTT connection",
     )
+    argparser.add_argument(
+        "--gdo0-gpio-line-name",
+        type=str,
+        required=True,
+        # GPIO24 recommended at
+        # https://github.com/fphammerle/python-cc1101/tree/v2.7.3#wiring-raspberry-pi
+        help="Name of GPIO pin that CC1101's GDO0 pin is connected to."
+        " Run command `gpioinfo` to get a list of all available GPIO lines."
+        " Recommended: GPIO24",
+    )
     # https://github.com/fphammerle/wireless-sensor/blob/v0.3.0/wireless_sensor/_cli.py#L28
     argparser.add_argument(
         "--unlock-spi-device",
@@ -360,22 +290,23 @@ def _main() -> None:
         args.homeassistant_node_id
     ):
         raise ValueError(
-            "invalid home assistant node id {!r} (length >= 1, allowed characters: {})".format(
-                args.homeassistant_node_id,
-                # pylint: disable=protected-access; false positive
-                wireless_sensor_mqtt._homeassistant.NODE_ID_ALLOWED_CHARS,
-            )
-            + "\nchange argument of --homeassistant-node-id"
+            f"invalid home assistant node id {args.homeassistant_node_id!r}"
+            " (length >= 1, allowed characters:"
+            f" {wireless_sensor_mqtt._homeassistant.NODE_ID_ALLOWED_CHARS})"
+            "\nchange argument of --homeassistant-node-id"
+        )
+    asyncio.run(
+        _run(
+            mqtt_host=args.mqtt_host,
+            mqtt_port=mqtt_port,
+            mqtt_disable_tls=args.mqtt_disable_tls,
+            mqtt_username=args.mqtt_username,
+            mqtt_password=mqtt_password,
+            mqtt_topic_prefix=args.mqtt_topic_prefix,
+            homeassistant_discovery_prefix=args.homeassistant_discovery_prefix,
+            homeassistant_node_id=args.homeassistant_node_id,
+            mock_measurements=args.mock_measurements,
+            gdo0_gpio_line_name=args.gdo0_gpio_line_name.encode(),
+            unlock_spi_device=args.unlock_spi_device,
         )
-    _run(
-        mqtt_host=args.mqtt_host,
-        mqtt_port=mqtt_port,
-        mqtt_disable_tls=args.mqtt_disable_tls,
-        mqtt_username=args.mqtt_username,
-        mqtt_password=mqtt_password,
-        mqtt_topic_prefix=args.mqtt_topic_prefix,
-        homeassistant_discovery_prefix=args.homeassistant_discovery_prefix,
-        homeassistant_node_id=args.homeassistant_node_id,
-        mock_measurements=args.mock_measurements,
-        unlock_spi_device=args.unlock_spi_device,
     )
